@@ -953,9 +953,10 @@ p0macro (char *bp, FILE *infd, FILE *outfd)
    if (libout)
    {
       FILE *libfd;
-      char libfile[MAXPATHNAMESIZE];
+      char libfile[MAXPATHNAMESIZE*2];
 
-      sprintf (libfile, "%s%s.mac", liboutfile.dirname, mac->macname);
+      snprintf (libfile, sizeof(libfile), "%s%s.mac",
+      		liboutfile.dirname, mac->macname);
       if ((libfd = fopen (libfile, "w")) != NULL)
       {
          fprintf (libfd,"%-6.6s $MACRO ", mac->macname);
