@@ -70,6 +70,7 @@
 *   10/19/15   DGP   Fixed print relocation marker when ext+offset.
 *   11/18/15   DGP   Added inpseg flag.
 *                    Fixed dseg and common symbols in literal.
+*   11/17/22   DGP   Fixed IDT and TITL test.
 *	
 ***********************************************************************/
 
@@ -3150,7 +3151,7 @@ p2pop (OpCode *op, char *bp, FILE *outfd, FILE *lstfd)
 	 while (*bp != '\'' && *bp != '\n') bp++;
       }
       *bp = '\0';
-      if (strlen(idtbuf) > IDTSIZE)
+      if (strlen(cp) > IDTSIZE)
       {
 	 sprintf (errtmp, "Invalid IDT value: %s", cp);
 	 logerror (errtmp);
@@ -3379,6 +3380,12 @@ p2pop (OpCode *op, char *bp, FILE *outfd, FILE *lstfd)
 	 while (*bp != '\'' && *bp != '\n') bp++;
       }
       *bp = '\0';
+      if (strlen (cp) > TTLSIZE)
+      {
+	 sprintf (errtmp, "Invalid TITL value: %s", cp);
+	 logerror (errtmp);
+	 *(cp+TTLSIZE) = '\0';
+      }
       strcpy (ttlbuf, cp);
       printed = TRUE;
       break;
