@@ -7,10 +7,27 @@
 *   07/10/03   DGP   If external is absolute and val 0, change relocatable.
 *   07/23/03   DGP   Added EBCDIC->ASCII for IBM OS/390 cross assemble.
 *   08/13/03   DGP   Added XREF support.
-*   08/14/03   DGP   Added copyopen and copyclose.
+*   08/14/03   DGP   Added opencopy and closecopy.
 *   08/18/03   DGP   Added parenthesis to expressions.
 *   05/20/04   DGP   Added literal support.
 *   06/07/04   DGP   Added FLOAT and DOUBLE support.
+*   12/26/05   DGP   Added DSEG support.
+*   12/30/05   DGP   Changes SymNode to use flags.
+*   07/30/07   DGP   Change mktemp to mkstemp.
+*   03/16/09   DGP   Correct float/double literal
+*   03/27/09   DGP   Unified error reporting using logerror.
+*   06/29/09   DGP   Check for duplicate error in logerror.
+*   04/22/10   DGP   Strip single quotes in COPY directive.
+*   12/01/10   DGP   Added ILL_EXTERNAL error.
+*   12/07/10   DGP   Added checkexpr function.
+*   12/14/10   DGP   If opencopy fopen fails try lower caseing the name.
+*                    Added file number to error processing.
+*   12/16/10   DGP   Added a unified source read function readsource().
+*   12/16/10   DGP   Added a unified source write function writesource().
+*   01/06/11   DGP   Fixed cross reference issues.
+*   01/04/14   DGP   Fixed more cross reference issues.
+*   11/18/15   DGP   Initialize segvaltype and csegval.
+*   10/01/16   DGP   Try COPY open "as is" before adding incldir.
 *	
 ***********************************************************************/
 
@@ -19,6 +36,12 @@
 #include <ctype.h>
 #include <string.h>
 #include <time.h>
+#include <fcntl.h>
+#ifndef WIN32
+#include <unistd.h>
+#else
+#include <io.h>
+#endif
 
 #include "asmdef.h"
 #include "errors.h"
@@ -27,26 +50,36 @@ extern int pc;
 extern int symbolcount;
 extern int rightmargin;
 extern int linenum;
+extern int indseg;
 extern int genxref;
 extern int filenum;
-extern int xrefgened;
+extern int fileseq;
 extern int errcount;
 extern int errnum;
 extern int absolute;
 extern int pgmlength;
 extern int codelength;
+extern int p1errcnt;
+extern int inpass;
+extern int txmiramode;
+extern int segvaltype;
+extern int csegval;
 extern char inbuf[MAXLINE];
 extern char incldir[MAXLINE];
+extern char *pscanbuf;
 extern SymNode *symbols[MAXSYMBOLS];
 extern char errline[10][120];
 extern COPYfile files[MAXASMFILES];
+extern ErrTable p1error[MAXERRORS];
+
+static int memexpr;
 
 #if defined(USS) || defined(OS390)
 /*
 ** EBCDIC to ASCII conversion table. 
 */
 
-static unsigned char ebcasc[256] =
+unsigned char ebcasc[256] =
 {
  /*00  NU    SH    SX    EX    PF    HT    LC    DL */
       0x00, 0x01, 0x02, 0x03, 0x3F, 0x09, 0x3F, 0x7F,
@@ -113,14 +146,120 @@ static unsigned char ebcasc[256] =
  /*F8   8     9                                     */
       0x38, 0x39, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0xFF
 };
-
-#define TOASCII(a) ebcasc[(a)]
-
-#else
-#define TOASCII(a) (a)
 #endif
 
 
+/***********************************************************************
+* readsource - Read source mode, line number and text from the temp file.
+***********************************************************************/
+
+int
+readsource (FILE *fd, int *inmode, int *linenum, int *filenum, char *inbuf)
+{
+#ifdef WIN32
+   int i;
+   char temp[32];
+#endif
+#if defined(DEBUGP1RDR) || defined(DEBUGP2RDR)
+   int flag = FALSE;
+#endif
+
+#ifdef WIN32
+   /*
+   ** Stupid WinBLOWS can't handle inter mixed binary/ascii reads....
+   ** Even though we fdopen in binary mode.
+   */
+
+   for (i = 0; i < 8; i++)
+   {
+      if ((temp[i] = fgetc(fd)) == EOF)
+         return (-1);
+   }
+   temp[8] = 0;
+   sscanf (temp, "%x", inmode);
+   for (i = 0; i < 8; i++)
+   {
+      if ((temp[i] = fgetc(fd)) == EOF)
+         return (-1);
+   }
+   temp[8] = 0;
+   sscanf (temp, "%d", linenum);
+#else
+   if (fread(inmode, 1, 4, fd) != 4)
+      return (-1);
+   if (fread(linenum, 1, 4, fd) != 4)
+      return (-1);
+#endif
+
+   if (fgets(inbuf, MAXLINE, fd) == NULL)
+      return (-1);
+
+   *filenum = (*inmode & NESTMASK) >> NESTSHFT;
+
+#ifdef DEBUGP1RDR
+   if (inpass == 1) flag = TRUE;
+#endif
+#ifdef DEBUGP2RDR
+   if (inpass == 2) flag = TRUE;
+#endif
+#if defined(DEBUGP1RDR) || defined(DEBUGP2RDR)
+   if (flag)
+   {
+      printf ("linenum = %d, inmode = %06x, filenum = %d\n",
+	      *linenum, *inmode, *filenum);
+      printf ("inbuf = %s", inbuf);
+   }
+#endif
+   return (0);
+}
+
+/***********************************************************************
+* writesource - Write source to the temp file.
+***********************************************************************/
+
+void
+writesource (FILE *fd, int mode, int num, char *buf, char *loc)
+{
+#ifdef WIN32
+   char temp[16];
+
+   /*
+   ** Stupid WinBLOWS can't handle inter mixed binary/ascii reads....
+   ** Even though we fdopen in binary mode.
+   */
+
+   sprintf (temp, "%08x", mode);
+   fwrite (temp, 1, 8, fd);
+   sprintf (temp, "%08d", num);
+   fwrite (temp, 1, 8, fd);
+#else
+   fwrite (&mode, 1, 4, fd);
+   fwrite (&num, 1, 4, fd);
+#endif
+
+   fputs (buf, fd);
+#ifdef DEBUGP0OUT
+   printf ("%s(%06X) %04d:%s", loc, mode, num, buf);
+#endif
+}
+
+#ifdef WIN32
+/***********************************************************************
+* mkstemp - Brain dead support for WinBlows.
+***********************************************************************/
+
+int mkstemp (char *template)
+{
+   int fd;
+   char tmpnam[1024];
+
+   strcpy (tmpnam, mktemp(template));
+   fd = open (tmpnam, O_CREAT | O_TRUNC | O_RDWR, 0666);
+   strcpy (template, tmpnam);
+   return (fd);
+}
+#endif
+
 /***********************************************************************
 * opencopy - Open a COPY
 ***********************************************************************/
@@ -130,56 +269,84 @@ opencopy (char *bp, FILE *tmpfd)
 {
    FILE *infd;
    char *cp;
-   char mkname[32];
+   int tmpdes;
    char tname[1024];
 
    while (isspace (*bp)) bp++;
    cp = bp;
+   if (*cp == '\'') cp++;
    while (!isspace (*bp)) bp++;
    *bp = '\0';
-   if (incldir[0])
-      sprintf (tname, "%s/%s", incldir, cp);
-   else
-      strcpy (tname, cp);
+   if (*(bp-1) == '\'') *(bp-1) = '\0';
+   strcpy (tname, cp);
    if ((infd = fopen (tname, "r")) == NULL)
    {
-      sprintf (tname, "asm990: Can't open input file for COPY: %s", cp);
-      perror (tname);
-      exit (1);
+      if (incldir[0])
+	 sprintf (tname, "%s/%s", incldir, cp);
+      if ((infd = fopen (tname, "r")) == NULL)
+      {
+	 if (isupper(*cp))
+	 {
+	    char *ccp;
+	    for (ccp = cp; *ccp; ccp++)
+	       if (isupper(*ccp)) *ccp = tolower (*ccp);
+	    if (incldir[0])
+	       sprintf (tname, "%s/%s", incldir, cp);
+	    else
+	       strcpy (tname, cp);
+	    infd = fopen (tname, "r");
+	 }
+      }
+      if (infd == NULL)
+      {
+	 sprintf (tname, "asm990: Can't open input file for COPY: %s", cp);
+	 perror (tname);
+	 exit (ABORT);
+      }
    }
    files[filenum].fd = tmpfd;
    files[filenum].line = linenum;
+
+#ifdef WIN32
+   sprintf (tname, "%s", TEMPSPEC);
+#else
+   sprintf (tname, "/tmp/%s", TEMPSPEC);
+#endif
+   if ((tmpdes = mkstemp (tname)) < 0)
+   {
+      perror ("asm990: Can't mkstemp for COPY");
+      exit (ABORT);
+   }
+   strcpy (files[filenum].tmpnam, tname);
+
    filenum++;
+   fileseq++;
+
    if (filenum >= MAXASMFILES)
    {
       fprintf (stderr, "Too many COPY files\n");
-      exit (1);
+      exit (ABORT);
    }
    linenum = 0;
-   strcpy (mkname, TEMPSPEC);
-   if ((cp = (char *)mktemp (mkname)) == NULL)
-   {
-      perror ("asm990: Can't mktemp for COPY");
-      exit (1);
-   }
-
-#ifdef WIN32
-   sprintf (tname, "%s.s", cp);
-#else
-   sprintf (tname, "/tmp/%s.s", cp);
-#endif
-   if ((tmpfd = fopen (tname, "w+")) == NULL)
+   if ((tmpfd = fdopen (tmpdes, "w+")) == NULL)
    {
       perror ("asm990: Can't open temporary file for COPY");
-      exit (1);
+      exit (ABORT);
    }
-   detab (infd, tmpfd);
+   detab (infd, tmpfd, "copy");
    fclose (infd);
    if (fseek (tmpfd, 0, SEEK_SET) < 0)
    {
       perror ("asm990: Can't rewind temp file");
-      exit (1);
+      exit (ABORT);
    }
+
+   /*
+   ** Mark tmp file for deletion
+   */
+
+   unlink (tname);
+
    return (tmpfd);
 }
 
@@ -194,6 +361,7 @@ closecopy (FILE *tmpfd)
    {
       fclose (tmpfd);
       filenum--;
+      unlink (files[filenum].tmpnam);
       linenum = files[filenum].line;
       return (files[filenum].fd);
    }
@@ -208,6 +376,7 @@ void
 Parse_Error (int cause, int state, int defer)
 {
    char errorstring[256];
+   char errtmp[256];
 
    if (defer) return;
    strcpy (errorstring, "Unknown error");
@@ -232,10 +401,22 @@ Parse_Error (int cause, int state, int defer)
    }
    else if (cause == INTERP_ERROR)
    {
-      strcpy (errorstring, "Divide by zero");
+      switch (state)
+      {
+      case ZERO_DIVIDE:
+	 strcpy (errorstring, "Divide by zero");
+	 break;
+      case ILL_EXTERNAL:
+	 strcpy (errorstring, "Illegal use of external variable");
+	 break;
+      case INTERNAL_ERR:
+	 strcpy (errorstring, "Internal Error");
+	 break;
+      default: ;
+      }
    }
-   errcount++;
-   sprintf (errline[errnum++], "%s in expression", errorstring);
+   sprintf (errtmp, "%s in expression", errorstring);
+   logerror (errtmp);
 
 } /* Parse_Error */
 
@@ -261,7 +442,7 @@ tokscan (char *buf, char **token, int *tokentype, int *tokenvalue, char *term)
       ** We stop scanning on a newline or when we're passed the operands.
       */
 
-      if (*buf == '\n' || (buf - inbuf > rightmargin))
+      if (*buf == '\n' || (buf - pscanbuf > rightmargin))
       {
 	 t[0] = '\0';
 	 *token = t;
@@ -304,7 +485,7 @@ tokscan (char *buf, char **token, int *tokentype, int *tokenvalue, char *term)
 ***********************************************************************/
 
 SymNode *
-symlookup (char *sym, int add, int noxref)
+symlookup (char *sym, int add, int xref)
 {
    SymNode *ret = NULL;
    int done = FALSE;
@@ -322,6 +503,7 @@ symlookup (char *sym, int add, int noxref)
    ** Empty symbol table.
    */
 
+   if (txmiramode) sym[MAXSHORTSYMLEN] = 0;
    if (symbolcount == 0)
    {
       if (!add) return (NULL);
@@ -339,12 +521,16 @@ symlookup (char *sym, int add, int noxref)
       symbols[0]->value = pc;
       symbols[0]->line = linenum;
 #ifdef DEBUGXREF
-      printf ("symlookup: sym = %s defline = %d\n", sym, linenum);
+      if (genxref && linenum)
+	 printf ("symlookup-1: sym = %s defline = %d, inpass = %d\n",
+		 sym, linenum, inpass);
 #endif
-      if (absolute)
-	 symbols[0]->relocatable = FALSE;
-      else
-	 symbols[0]->relocatable = TRUE;
+      if (indseg)
+	 symbols[0]->flags = RELOCATABLE | DSEGSYM;
+      else if (!absolute)
+	 symbols[0]->flags = RELOCATABLE;
+      if (!txmiramode && ((strlen (sym) > MAXSHORTSYMLEN) || (sym[0] == '_')))
+	 symbols[0]->flags |= LONGSYM;
       symbolcount++;
       return (symbols[0]);
    }
@@ -369,33 +555,72 @@ symlookup (char *sym, int add, int noxref)
       {
 	 SymNode *symp = symbols[mid];
 
-	 if (add && !symp->predefine)
+	 if (add && !(symp->flags & PREDEFINE))
 	    return (NULL); /* must be a duplicate */
 
-	 if (genxref && !noxref)
+	 if (genxref && xref && (inpass != 1))
 	 {
 	    XrefNode *new;
+	    XrefNode *curr, *last;
+	    int found = FALSE;
 
-	    if ((new = (XrefNode *)malloc (sizeof (XrefNode))) == NULL)
+	    curr = symp->xref_head;
+	    last = NULL;
+	    while (curr)
 	    {
-	       fprintf (stderr, "asm990: Unable to allocate memory\n");
-	       exit (ABORT);
+	       if (curr->line == linenum)
+	       {
+	          found = TRUE;
+		  break;
+	       }
+	       if (curr->line > linenum)
+	          break;
+	       last = curr;
+	       curr = curr->next;
 	    }
+	    if (!found)
+	    {
+	       if ((new = (XrefNode *)malloc (sizeof (XrefNode))) == NULL)
+	       {
+		  fprintf (stderr, "asm990: Unable to allocate memory\n");
+		  exit (ABORT);
+	       }
 #ifdef DEBUGXREF
-	    printf ("symlookup: sym = %s refline = %d\n", sym, linenum);
+	       if (linenum)
+		  printf ("symlookup-2: sym = %s defline = %d, inpass = %d\n",
+			  sym, linenum, inpass);
 #endif
-	    new->next = NULL;
-	    new->line = linenum;
-	    if (symp->xref_head == NULL)
-	    {
-	       symp->xref_head = new;
+	       new->next = NULL;
+	       new->line = linenum;
+	       if (symp->xref_head == NULL)
+	       {
+		  symp->xref_head = new;
+		  symp->xref_tail = new;
+	       }
+	       else if (curr)
+	       {
+		  if (last)
+		  {
+		     new->next = last->next;
+		     last->next = new;
+		  }
+		  else
+		  {
+		     new->next = symp->xref_head;
+		     symp->xref_head = new;
+		  }
+	       }
+	       else
+	       {
+		  (symp->xref_tail)->next = new;
+		  symp->xref_tail = new;
+	       }
 	    }
-	    else
-	    {
-	       (symp->xref_tail)->next = new;
-	    }
-	    symp->xref_tail = new;
 	 }
+#ifdef DEBUGSYM
+	 printf (" sym found: val = %04X, flags = %X\n",
+	 	 symp->value, symp->flags);
+#endif
          return (symbols[mid]);
       }
       else if (r < 0)
@@ -435,14 +660,18 @@ symlookup (char *sym, int add, int noxref)
       memset (new, '\0', sizeof (SymNode));
       strcpy (new->symbol, sym);
 #ifdef DEBUGXREF
-      printf ("symlookup: sym = %s defline = %d\n", sym, linenum);
+      if (genxref && linenum)
+	 printf ("symlookup-3: sym = %s defline = %d, inpass = %d\n",
+		 sym, linenum, inpass);
 #endif
       new->value = pc;
       new->line = linenum;
-      if (absolute)
-	 new->relocatable = FALSE;
-      else
-	 new->relocatable = TRUE;
+      if (indseg)
+	 new->flags = RELOCATABLE | DSEGSYM;
+      else if (!absolute)
+	 new->flags = RELOCATABLE;
+      if (!txmiramode && ((strlen (sym) > MAXSHORTSYMLEN) || (sym[0] == '_')))
+	 new->flags |= LONGSYM;
 
       /*
       ** Insert pointer in sort order.
@@ -479,6 +708,7 @@ exprscan (char *bp, int *val, char *tterm, int *relo, int entsize, int defer,
    int index = 0;
    int lrelo = 0;
    tokval lval = 0;
+   char errtmp[MAXLINE];
 
 #ifdef DEBUGEXPR
    printf ("exprscan: entered: bp = %s\n", bp);
@@ -488,44 +718,49 @@ exprscan (char *bp, int *val, char *tterm, int *relo, int entsize, int defer,
 #ifdef DEBUGEXPR
    printf ("   bp = %s\n", bp);
 #endif
+   segvaltype = 0;
+   csegval = 0;
    if (*bp == LITERALSYM)
    {
-      SymNode *s;
-      char *cp;
-      char temp[32];
+      if (!txmiramode)
+      {
+	 SymNode *s;
+	 char *cp;
+	 char temp[32];
 
-      cp = bp + 1;
-      while (*cp && !isspace(*cp) && *cp != ',') cp++;
-      index = cp - bp;
-      strncpy (temp, bp, index);
-      temp[index] = '\0';
+	 cp = bp + 1;
+	 while (*cp && !isspace(*cp) && *cp != ',') cp++;
+	 index = cp - bp;
+	 strncpy (temp, bp, index);
+	 temp[index] = '\0';
 
-#ifdef DEBUGLITERAL
-      printf ("exprscan: Literal: temp = %s\n", temp);
+#if defined(DEBUGLITERAL) || defined(DEBUGEXTNDX)
+	 printf ("exprscan: Literal: pass = %d, temp = %s\n", inpass, temp);
 #endif
 
-      if ((s = symlookup (temp, FALSE, FALSE)) == NULL)
-      {
-	 if ((s = symlookup (temp, TRUE, FALSE)) == NULL)
+	 if ((s = symlookup (temp, FALSE, TRUE)) == NULL)
 	 {
-	    errcount++;
-	    sprintf (errline[errnum++], "Unable to add Literal: %s", temp);
-	    lval = 0;
+	    if ((s = symlookup (temp, TRUE, TRUE)) == NULL)
+	    {
+	       sprintf (errtmp, "Unable to add Literal: %s", temp);
+	       logerror (errtmp);
+	       lval = 0;
+	    }
+	    s->value = codelength;
+	    if (temp[1] == 'F' || temp[1] == 'L')
+	       codelength += 4;
+	    else if (temp[1] == 'D' || temp[1] == 'Q')
+	       codelength += 8;
+	    else
+	       codelength += 2;
 	 }
-	 s->value = codelength;
-	 if (temp[1] == 'E' || temp[1] == 'L')
-	    codelength += 4;
-	 else if (temp[1] == 'D' || temp[1] == 'Q')
-	    codelength += 8;
-	 else
-	    codelength += 2;
+	 lval = s->value;
+	 if (s->flags & RELOCATABLE) lrelo++;
       }
-      lval = s->value;
-      if (s->relocatable) lrelo++;
    }
    else if (strlen(bp) && *bp != ',')
    {
-      lval = Parser (bp, &lrelo, &index, entsize, defer, pcinc);
+      lval = Parser (bp, &lrelo, &memexpr, &index, entsize, defer, pcinc);
    }
 #ifdef DEBUGEXPR
    printf (" index = %d, *(bp+index) = %02X\n",  index, *(bp+index));
@@ -591,3 +826,122 @@ cvtdouble (char *bp)
 #endif
    return (tv);
 }
+
+/***********************************************************************
+* logerror - Log an error for printing.
+***********************************************************************/
+
+void
+logerror (char *errmsg)
+{
+   int i;
+#ifdef DEBUGERROR
+   printf ("logerror: inpass = %d, linenum = %d, errcount = %d\n",
+	   inpass, linenum, errcount);
+   printf ("   p1errcnt = %d, errnum = %d, errorfile = %d\n",
+	   p1errcnt, errnum, (filenum ? fileseq : 0));
+   printf ("   errmsg = %s\n", errmsg);
+#endif
+
+   if (inpass != 2)
+      for (i = 0; i < p1errcnt; i++)
+      {
+	 if ((p1error[i].errorline == linenum) && 
+	     (p1error[p1errcnt].errorfile == (filenum ? fileseq : 0)) &&
+	     !strcmp(p1error[i].errortext, errmsg))
+	 {
+#ifdef DEBUGERROR
+	    printf ("   DUPLICATE\n");
+#endif
+	    return;
+	 }
+      }
+
+   errcount++;
+   if (inpass == 2)
+   {
+      strcpy (errline[errnum++], errmsg);
+   }
+   else
+   {
+#ifdef DEBUGERROR
+      printf ("   NEW, queue\n");
+#endif
+      p1error[p1errcnt].errortext = (char *)malloc (120);
+      p1error[p1errcnt].errorline = linenum;
+      p1error[p1errcnt].errorfile = (filenum ? fileseq : 0);
+      strcpy (p1error[p1errcnt].errortext, errmsg);
+      p1errcnt++;
+   }
+}
+
+/***********************************************************************
+* checkexpr - Check expression.
+***********************************************************************/
+
+char *
+checkexpr (char *bp, int *type)
+{
+   char *cp;
+   int val, relo;
+   char term;
+   char tmpexpr[256];
+
+   *type = 0;
+   cp = bp;
+   while (*cp && !isspace(*cp) && *cp != ',') cp++;
+   val = cp - bp;
+   strncpy (tmpexpr, bp, val);
+   tmpexpr[val] = '\0';
+
+#ifdef DEBUGCHECKEXPR
+   printf ("checkexpr: expr = %s\n", tmpexpr);
+#endif
+
+   if (tmpexpr[--val] == ')')
+   {
+      tmpexpr[val--] = '\0';
+      while (val && tmpexpr[val] != '(') val--;
+      tmpexpr[val] = '\0';
+      if (val)
+      {
+#ifdef DEBUGCHECKEXPR
+	 printf ("   Indexed Memory expression\n");
+#endif
+	 *type = MEMEXPR;
+	 return cp;
+      }
+   }
+
+   if (*bp == MEMSYM)
+   {
+#ifdef DEBUGCHECKEXPR
+      printf ("   Memory expression\n");
+#endif
+      *type = MEMEXPR;
+      return cp;
+   }
+   else if (*bp == LITERALSYM)
+   {
+#ifdef DEBUGCHECKEXPR
+      printf ("   Literal expression\n");
+#endif
+      *type = LITEXPR;
+      return cp;
+   }
+
+   exprscan (tmpexpr, &val, &term, &relo, 2, TRUE, 0);
+   if (relo || memexpr || val > 15)
+   {
+#ifdef DEBUGCHECKEXPR
+      printf ("   Memory expression\n");
+#endif
+      *type = MEMEXPR;
+      return cp;
+   }
+#ifdef DEBUGCHECKEXPR
+   printf ("   Register expression\n");
+#endif
+   *type = REGEXPR;
+   return cp;
+}
